#include <stdio.h>
#include <string.h>
#include <time.h>

/* costanti generali */
#define FALSE 0
#define TRUE 1
#define OK 0
#define ERROR 1

/* costanti versione */
#define TITLE "xprime"
#define VERSION "v0.12.05082002"
#define YEAR "2002"

/* file */
#define CONFIGNAME "config.xpr"
#define EXTNAME ".xpr"
#define TEXTNAME ".txt"

/* debug */
#define ALL      4  // frequente
#define LOADPAGE 3
#define GOTPRIME 2
#define SAVEPAGE 1  // raro
#define NODEBUG  0
#define FREQOUT freqout
#define INVALIDDEBUG "BADDBUG"

char *dbstr[5];


/* vars */
clock_t clockEnd, clockStart;
FILE *f;
long cfgPageSize;          // lunghezza di pagina da config file o parametri
long pageSize = 0;         // lunghezza di ogni pagina
int freqout;              // debug, ogni quanto stampare
char *uniqueName;         // nome del file unico da creare
char *prefixName;         // prefisso nome pagine
char *ctrlname;           // nome del file control
int create;               // flag, true se deve creare file unico
int interact;             // flag per l'interattivita'
long timeout;             // tempo timeout in secondi
long startindex;          // indice del primo primo della pagina corrente
long size;                // grandezza (variabile) della ultima pagina
//int start;                // flag: se e' il primo run


long *loadedPage,         // pagina caricata
     *curPage;            // pagina in fase di elaborazione
long iloadPage = 0,        // index pagina caricata
    icurPage = 0;            // indice della apgina corrente
long pageIndex,           // indice del primo nella pagina
     maxPrime;            // massimo primo trovato

int isPrime;
char name[30],
     date[30];

/* livello di debug */
int debug = GOTPRIME;

void saveControl(int start, int toZero);
void loadPage(long page, long *startindex, long *size);
void savePage(long page, long startindex, long size);
long getPrime(long i);
long pageSetPrime(long i);


double timer() {
 clock_t t = clock();

 return ((double)t)/CLOCKS_PER_SEC;
}

/*
trasforma numero in stringa.
[NB] la lunghezza massima e' fissata
*/
#define MAXLEN 10
char *toString(int num, int len){
  char buff[MAXLEN];
  int i;
  buff[len] = 0;

  for (i=len-1; i>=0; i--){
    buff[i] = num%10 + 48;
    num = num/10;
  }

  return strdup(buff);
}

void startControl() {
     saveControl(TRUE, TRUE);
}

/*
carica file control
nome, anno, massimo dei primi, numero dei file
*/
int loadControl() {
  char string[30];

  /* controlla esistenza file controllo */
  f = fopen(ctrlname, "r");

  /* se il file non esiste */
  if (f != NULL) {
     fscanf(f,"%s %s", name, date); // intro
     fscanf(f, "%s %D ", string, &maxPrime); /* massimo tra i primi */
     fscanf(f, "%s %D ", string, &icurPage); /* numero di page */
     fscanf(f, "%s %D ", string, &pageSize); /* page size */
     fscanf(f, "%s %D ", string, &size); /* size ultima page */
     fclose(f);

     if (debug >= LOADPAGE) {
       printf(" Control loaded: max of prime=%D", maxPrime);
       printf(" page number=%D\n\r", icurPage);
     }
     return TRUE;
  } else
     return FALSE;
}

char *debugName(int debug) {
 if ((debug >= 0) && (debug <=ALL)) {
   return dbstr[debug];
 } else {
   return INVALIDDEBUG;
 }
}

char *trueStr(int flag) {
  if (flag)
     return "true";
  else
     return "false";
}
/*
carica file config
*/
void loadConfig() { 

  char string[30];
  /* valori di default */
  cfgPageSize = 100;
  prefixName = strdup("prime");
  freqout = 5;
  debug = ALL;
  interact = FALSE;
  /* [TODO] free di questa stringa */
  uniqueName = strdup("unique.txt");
  create = FALSE;
  timeout = 60;

  /* controlla esistenza file controllo */
  f = fopen(CONFIGNAME, "r");

  /* se il file esiste */
  if (f != NULL) {     
     fscanf(f, "%s %D ", string, &cfgPageSize); /* pageSize */
     fscanf(f, "%s %s ", string, name);       /* prefix name */
     prefixName = strdup(name);
     fscanf(f, "%s %d ", string, &freqout);  /* freqout */
     fscanf(f, "%s %d ", string, &debug);    /* debug */
     fscanf(f, "%s %d ", string, &interact); /* interact */
     fscanf(f, "%s %s ", string, name);       /* unique name */
     uniqueName = strdup(name);   
     fscanf(f, "%s %d ", string, &create);   /* create */
     fscanf(f, "%s %D ", string, &timeout);   /* timeout */
  } else {
     printf("WARNING: config file (%s)not exist\n\r", CONFIGNAME);
  }
  fclose(f);

    printf("Config loaded ");
    printf("page size:%D, prefix:%s, frequency out:%d\r\n", 
            pageSize, prefixName, freqout);
    printf("debug state:%s, interact state:%s, unique name:%s\r\n",
	     debugName(debug), trueStr(interact), uniqueName);
    printf("create flag:%s, time out sec:%D\r\n",
            trueStr(create), timeout);
}

/*
salva il file control
se size < 0 maxprimi = 1, pone size = 0
*/
void saveControl(int start, int toZero) {
  /* controlla esistenza file controllo */
  f = fopen(ctrlname, "w");

  fprintf(f, "%s %s \r\n\r\n", TITLE, VERSION);
  if (start)  {
    fprintf(f, "maxprimi: %D \r\n", 1);         /* massimo dei primi */
  } else
     fprintf(f, "maxprimi: %D \r\n", curPage[size - 1]);

  fprintf(f, "lastpage:  %D \r\n", icurPage ); /* indice pagina finale */
  fprintf(f, "pageSize: %D \r\n", pageSize);   /* pageSize */

  if (toZero)
    size = 0;
  fprintf(f, "lastSize: %D \r\n", size);         /* size ultima page */

  fclose(f);

  if (debug >= SAVEPAGE)

     printf("*Save page #%D maxPrime=%D, sec %f\r\n",
	icurPage,
	curPage[pageSize - 1], 
	timer());
}

void help() {
    printf("usage  : xprime -ver  \r\n");
    printf("         xprime --help \r\n");
    printf("         xprime G -[c name|d val|f val|i|m val|p name|f val]]* \r\n");
}

/* copia le pagine */
/* @param pageSize che e' una variabile globale */
void copyPage(long *to, long *from) {
  long i;
  for (i=0; i<pageSize; i++)
    to[i] = from[i];
}

int main(int argc, char *args[]) {
  int i;

  char name[30];

  dbstr[SAVEPAGE] = "SAVEPAGE";
  dbstr[GOTPRIME] = "GOTPRIME";
  dbstr[LOADPAGE] = "LOADPAGE";
  dbstr[ALL] = "ALL";
  dbstr[NODEBUG] = "NODEBUG";
  loadConfig(); // carica valori di default

  name[0]=0;
  strcpy(name, prefixName);
  strcat(name, ".ctl");

  ctrlname = strdup(name);

  if (argc == 1) {
    printf("Using default config (config.xpr) \r\n");
  } else
  /* parte per '-', quindi -ver o --help */
  if (args[1][0] == '-')  {
    if (!strcmp(args[1], "-ver")) {
      printf("   %s version %s\n\r",TITLE,VERSION);
    } else if (!strcmp(args[1], "--help")) {
      help();
    } else {
      printf("ERROR: Wrong place for command \"%s\", or not exist\n\r", args[1]);
      printf("       for help xprime --help\n\r");
      printf("       hint: start with (-ver, --help) or (R, D)\r\n");
    }
    return OK;

  /* comando G[o] */
  } else if (args[1][0] == 'G') {
    for ( i=2; i<argc; i++) {
      printf("DEBUG: go %s:%d\n\r", args[i], i);
      if (args[i][0] == '-') {
      switch(args[i][1]) { 
        case 'c': //create
          i++;
          free(uniqueName);
          uniqueName = strdup(args[i]);
          create = TRUE;
          if (uniqueName[0] == '-') {
             printf("ERROR: %s:%i not a name\n\r", uniqueName,i);
          }
          if (debug >= ALL)
            printf("create unique name %s\n\r", uniqueName);
        break;

	case 'd': //debug level
          i++;
          debug = atoi(args[i]);
          if (debug<0) {
            printf("ERROR: debug value must be >= 0\n\r");
            return OK;
          }
          if (debug >= ALL)
             printf("set debug level: %d\n\r", debug);
        break;

        case 'f': //freqout
          i++;
          freqout = atoi(args[i]);
          if (freqout<10 ) {
            printf("ERROR: freqout value must be > 10, (value:%d)\r\n", freqout);
	    return OK;
          }
          if (debug >= ALL)
             printf("set frequency out %d\n\r",freqout);
        break;

        case 'i': // interattivita' on
          interact = TRUE;
          if (debug >= ALL)
             printf("set interactivity\n\r");
        break;

        case 'm': //pageSize
          i++;
          cfgPageSize = atoi(args[i]);
          if (pageSize <= 1) {
             printf("ERROR: pageSize must be > 1 (value:%D)\r\n", pageSize);
             return OK;
          }
          if (debug >= ALL)
             printf("set pageSize to %D\n\r", pageSize);
        break;

        case 'p': //prefix
          i++;
          free(prefixName);
          prefixName = strdup(args[i]);
          if (prefixName[0] == '-') {
             printf("ERROR: %s not a name\n\r", prefixName);
          }
          if (debug >= ALL)
             printf("set prefix %s\n\r", prefixName);
        break;

        case 't': //timeout in minuti
          i++;
          timeout = atoi(args[i]) * 60; // in secondi
          if (debug >= ALL)
             printf("set time out %D\n\r",timeout);
        break;

        default:
          printf("ERROR: command doesn't exist %s:%i\n\r", args[i],i);
          return OK;
        break;
      }
      } else {
        printf("ERROR: %s:%i not a command\n\r", args[i], i);
      }
    }
  } else  {
      printf("ERROR: Bad command \"%s\", for help xprime --help\n\r", args[1]);
      return OK;
  }



  if (create) {
    printf("create uniquefile %s, [NOT IMPLEMENTED]", uniqueName);
    exit(0);
  }

  if (loadControl()) {
    /* se si e' deciso di cambiare la grandezza della pagina */
    /* cambia il prefisso test --> r_test */
    /* carica l'i-esimo primo delle pagine test*/
      /* sinche' non e' pieno una pagina r_test */
      /* oppure sono finiti i primi test */
         /* salva pagina */
    if (cfgPageSize != pageSize) {
      printf("NOT IMPLEMENTED page size change (%D --> %D)",
            pageSize,cfgPageSize);
      exit(0);
    }
    //start = FALSE;
  } else {
    if (cfgPageSize <= 0) {
      printf("ERROR: pagesize must be > 0 (%D)", cfgPageSize);
      exit(0);
    }
    pageSize = cfgPageSize;
    size = 0;
    startControl();
  }
  loadedPage = (long*) malloc(pageSize * sizeof(long));
  curPage = (long*)malloc(pageSize * sizeof(long));

  if (freqout == 0) {
    printf("ERROR: freqout must be > 0 (%d)", freqout);
    exit(0);
  }
  if (freqout > pageSize) {
    printf("ERROR: freqout(%d) must be <= pageSize(%D)\r\n",freqout,pageSize);
    return OK;
  }

  while (1) {
    doPage();

    if (interact) {
      printf("interactive stop [NOT IMPLEMENTED]");
      exit(0);
    }

  }
  printf("done, total time %f, number of primes %D \r\n", timer(),
                      icurPage * pageSize);
  return 0;
}

/* calcola e' salva la prossima pagina*/
int doPage() {
  long idiv; /* indice dei divisori*/
  long div; /* divisore */
  long prime; /* numero da testare */
  long buf;
  long loadPageSize;

  loadControl();


  /* l'ultima pagina non e' stata completata*/
  /* size = 0 pagina completata */
  if (size != 0) {
    loadPage(icurPage, &startindex, &loadPageSize);
    copyPage(curPage, loadedPage);
  } else
    icurPage++;

  /* init page */
  if (maxPrime == 1){
    icurPage = 0;
    curPage[0] = 2;
  } else {
    iloadPage = 0;
    loadPage(iloadPage,&startindex,&loadPageSize);
  }

  /* per ogni elemento in questa pagina*/
  for ( pageIndex=size, prime=maxPrime+1; 
        pageIndex<pageSize; 
        prime++) 
  {
    isPrime = TRUE;
    for ( idiv = 0; getPrime(idiv) <= prime/2; idiv++ ){
      div = getPrime(idiv);
      buf = prime/div;
      if ( buf*div == prime ){
         isPrime = FALSE;
         break;
      }
    }

    if(isPrime) {
      pageSetPrime(prime);
      pageIndex++;

      size = pageIndex; /* numero di primi in questa pagina */

      if (timer() > timeout) {
        printf("time out:%D (page:%D, num prime:%D)\r\n",
                timeout, icurPage, pageIndex+pageSize*icurPage);
        savePage(icurPage, pageSize*icurPage, pageIndex);
        saveControl(FALSE, FALSE);
        
        exit(0);
      }
    }
  }

  savePage(icurPage, startindex, pageSize);
  saveControl(FALSE, TRUE);
  return 0;
}

long getPrime(long i) {
  long reqpage;   // pagina richiesta

  reqpage = i/pageSize;

  /* non e' la pagina in elaborazione */
  if (reqpage != icurPage) {
    /* non e' la pagina gia' caricata */
    if (reqpage != iloadPage) {
      iloadPage = reqpage;
      loadPage(iloadPage,&startindex,&size);
    }
    return loadedPage[i - iloadPage * pageSize];
  } else {
    return curPage[i- icurPage * pageSize];
  }
}

long pageSetPrime(long i) {
  curPage[pageIndex] = i;

  /* debug */
  if (debug >= GOTPRIME && (pageIndex%(pageSize/FREQOUT) == 0)) {
     clockEnd = timer();
     printf("  *primo %D, primeIndex %D, number %D, sec %f\r\n", i,
           pageIndex, pageIndex + icurPage*pageSize, timer());
     clockStart = timer();
  }
}

/* carica la pagina i-esima  */
/* dentro la pagina e' scritta la lunghezza della stessa */
void loadPage(long page, long *startindex, long *size) {
  int i;
  long got;
  char fn[30];
  fn[0]= 0;

  strcat(fn, prefixName);
  strcat(fn, toString(page, 4));
  strcat(fn, EXTNAME);

  f = fopen(fn,"r");
  if (f == NULL) {
    printf("ERROR: page \"%s\" do not exist", fn);
    exit(OK);
  }
  fscanf(f,"%s %D ", fn, startindex);
  fscanf(f,"%s %D ", fn, size);

  for (i = 0; i<*size; i++){
    fscanf(f, "%D", &got);
    loadedPage[i] = got;
  }
  fclose(f);
  if (debug >= LOADPAGE)
    printf("    *load page %D, file name %s\r\n",page, fn) ;
}

void savePage(long page, long startindex, long size) {
  int i;
  long got;
  char fn[30];
  fn[0]= 0;

  strcat(fn, prefixName);
  strcat(fn, toString(page, 4));
  strcat(fn, EXTNAME);

  f = fopen(fn,"w");
  fprintf(f,"startindex %D \r\n", startindex);
  fprintf(f,"size %D \r\n", size);

  for (i = 0; i<size; i++) {
    fprintf(f,"%10D ", curPage[i]);
    if (i % 5 == 0)
       fprintf(f,"\r\n");
  }
  fclose(f);
}
